/* convert FDIOUSER style pointlist to extended St. Louis style
	format and merge it with the nodelist (region segment)
	and/or produce a FDPOINT.PVT compatible pointlist.
	(c) 1993 - 1997, Michiel van der Vlist  */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>
#include <io.h>
#include <conio.h>
#include <time.h>
#include <fcntl.h>

/* ******************************************************************* */
/* Stuff below is meant to compile for OS/2, using Watcom C/C++ v10.0. */
/* It's just a Q&D hack to make things work, nothing is optimized for  */
/* OS/2, no functionality changes. But it seems to work :-)            */
/* This and all other #ifdefs with __OS2__ and __WATCOMC__ are for     */
/* OS/2 & Watcom, and:                                                 */
/* Made by Gerard van Essen, 2:281/527, March 26 1995                  */
/*                                                                     */
/* Build an OS/2 executable using Watcom C/C++ v10.0 with:             */
/* wcl386 /wx /x /j /olrs /mf /4r pcmerge                              */
/* ******************************************************************* */

#if defined (__WATCOMC__) && defined (__OS2__)

  /* Watcom has header files with different names */

  #include <malloc.h>
  #include <direct.h>

  /* Required to use OS/2's DosBeep() */

  #define INCL_DOSPROCESS
  #include <os2.h>

  /* Watcom uses other constants */

  #define MAXDRIVE _MAX_DRIVE
  #define MAXDIR   _MAX_DIR
  #define MAXFILE  _MAX_FNAME
  #define MAXEXT   _MAX_EXT
  #define MAXPATH  _MAX_PATH

  /* OS/2 doesn't need the far/huge stuff */

  #undef far
  #undef huge

  #define far
  #define huge

  #define farcalloc calloc
  #define farfree   free

  // So we don't need special copy/compare functions either.

  #define Strcpy(a,b) strcpy((char *)(a), (char *)(b))
  #define strcpY(a,b) strcpy((char *)(a), (char *)(b))
  #define Strcmp(a,b) strcmp((char *)(a), (char *)(b))

  /* Always report a nice amount of free memory for OS/2 */

  #define farcoreleft() 750000L

  /* Some other differences */

  #define findfirst(a, b, c) _dos_findfirst(a, c, b)
  #define findnext  _dos_findnext

  #define EXTENSION 0x02
  #define FILENAME  0x04
  #define DIRECTORY 0x08
  #define DRIVE     0x10

  int fnsplit(const char *path, char *drive, char *dir, char *name, char *ext);
  #define fnmerge _makepath

  /* Then the inevitable date sh*t */

  struct date {
      int  da_year;
      char da_day;
      char da_mon;
  };

  void getdate(struct date *d);

#else

  #include <alloc.h>
  #include <dir.h>
  #include <values.h>

#endif

/* ******************************************************************* */


#define MYNAMEANDNUMBER "(c) 1997, Michiel van der Vlist, PA0MMV (2:500/9.5)"

#ifdef __OS2__
  #define VERSION "2.8/2"
#else
  #define VERSION "2.8"
  #define TRUE (1==1)
  #define FALSE (!TRUE)
#endif

#define DEFAULTZONE 2
#define DEFAULTREGION 28
#define MINENTRYS 2000
#define MAXENTRYS 25000
#define NAMELENG 40
#define AVERNAMELEN 14
#define AVERPOINTS 11

#define INPLISTN "POINTPL"
#define OUTPLISTN "POINTS"
#define OUTPLISTE ".LST"
#define REGFILE "REGION28"
#define FDFILEN "FDPOINTS"
#define FDFILEE ".LST"
#define PVTLISTN "MYPOINTS"
#define NRPRIVATS 20
#define LOGFILE "PCMERGE.LOG"
#define CONFIGFILE "PCMERGE.CFG"
#define MAXDAYS 8

typedef unsigned char boolean;
typedef unsigned char byte;
typedef char string[NAMELENG];
typedef char dstring[NAMELENG*2];
typedef char longstring[80];

typedef struct {	byte flags;
						char huge *namep;
						unsigned int zone,net,node,point;
						int bossnr;
					}	p_entry;

p_entry huge *p_list;
p_entry far *p_l1;
char huge *pname;
char far *p_name;
char huge *last_pname;
unsigned int index[MAXENTRYS], an, Defaultzone=DEFAULTZONE;
unsigned long totalnamechars=0;
int ap,nameclean=1,nrofdups=0, nrofskips=0, privats=-1;;
int nrofprivs=0, aantbosses, errornumber=0, eolerrors=0, maxeolerrors=10;
boolean sortflag=FALSE, mergeflag=TRUE, commentflag=TRUE, errabort=FALSE;
boolean fdflag=TRUE, logflag=FALSE, multiflag=0, zapflag=FALSE;
boolean npcflag=FALSE,  usepointname=FALSE, pwarning=FALSE;
boolean excludeorphan=FALSE;
byte usebossname=0, usebossloc=0, usebossfone=0;
FILE *infile,*outfile,*log;
fpos_t far *bosspos1;
fpos_t huge *bosspos;
longstring pointlist,spointlist,nodelist,outlist,fdlist,logfile;
char pstr[256];
longstring huge *pvtlist;
longstring far *pvt_list;
dstring pointphonenr="-Unpublished-", baud="2400,";
string location="_", sysname="_";
char separator='.', pext[5];

/* Partial remainder tables for CCIT16
	CRCC as used with REMOTE and XMODEM.
	Polynomial is x^16+x^12+x^5+1 */

unsigned const
char CRC1[256]={	0x00,0x10,0x20,0x30,0x40,0x50,0x60,0x70,
						0x81,0x91,0xA1,0xB1,0xC1,0xD1,0xE1,0xF1,
						0x12,0x02,0x32,0x22,0x52,0x42,0x72,0x62,
						0x93,0x83,0xB3,0xA3,0xD3,0xC3,0xF3,0xE3,
						0x24,0x34,0x04,0x14,0x64,0x74,0x44,0x54,
						0xA5,0xB5,0x85,0x95,0xE5,0xF5,0xC5,0xD5,
						0x36,0x26,0x16,0x06,0x76,0x66,0x56,0x46,
						0xB7,0xA7,0x97,0x87,0xF7,0xE7,0xD7,0xC7,
						0x48,0x58,0x68,0x78,0x08,0x18,0x28,0x38,
						0xC9,0xD9,0xE9,0xF9,0x89,0x99,0xA9,0xB9,
						0x5A,0x4A,0x7A,0x6A,0x1A,0x0A,0x3A,0x2A,
						0xDB,0xCB,0xFB,0xEB,0x9B,0x8B,0xBB,0xAB,
						0x6C,0x7C,0x4C,0x5C,0x2C,0x3C,0x0C,0x1C,
						0xED,0xFD,0xCD,0xDD,0xAD,0xBD,0x8D,0x9D,
						0x7E,0x6E,0x5E,0x4E,0x3E,0x2E,0x1E,0x0E,
						0xFF,0xEF,0xDF,0xCF,0xBF,0xAF,0x9F,0x8F,
						0x91,0x81,0xB1,0xA1,0xD1,0xC1,0xF1,0xE1,
						0x10,0x00,0x30,0x20,0x50,0x40,0x70,0x60,
						0x83,0x93,0xA3,0xB3,0xC3,0xD3,0xE3,0xF3,
						0x02,0x12,0x22,0x32,0x42,0x52,0x62,0x72,
						0xB5,0xA5,0x95,0x85,0xF5,0xE5,0xD5,0xC5,
						0x34,0x24,0x14,0x04,0x74,0x64,0x54,0x44,
						0xA7,0xB7,0x87,0x97,0xE7,0xF7,0xC7,0xD7,
						0x26,0x36,0x06,0x16,0x66,0x76,0x46,0x56,
						0xD9,0xC9,0xF9,0xE9,0x99,0x89,0xB9,0xA9,
						0x58,0x48,0x78,0x68,0x18,0x08,0x38,0x28,
						0xCB,0xDB,0xEB,0xFB,0x8B,0x9B,0xAB,0xBB,
						0x4A,0x5A,0x6A,0x7A,0x0A,0x1A,0x2A,0x3A,
						0xFD,0xED,0xDD,0xCD,0xBD,0xAD,0x9D,0x8D,
						0x7C,0x6C,0x5C,0x4C,0x3C,0x2C,0x1C,0x0C,
						0xEF,0xFF,0xCF,0xDF,0xAF,0xBF,0x8F,0x9F,
						0x6E,0x7E,0x4E,0x5E,0x2E,0x3E,0x0E,0x1E };
unsigned const
char CRC2[256]={	0x00,0x21,0x42,0x63,0x84,0xA5,0xC6,0xE7,
						0x08,0x29,0x4A,0x6B,0x8C,0xAD,0xCE,0xEF,
						0x31,0x10,0x73,0x52,0xB5,0x94,0xF7,0xD6,
						0x39,0x18,0x7B,0x5A,0xBD,0x9C,0xFF,0xDE,
						0x62,0x43,0x20,0x01,0xE6,0xC7,0xA4,0x85,
						0x6A,0x4B,0x28,0x09,0xEE,0xCF,0xAC,0x8D,
						0x53,0x72,0x11,0x30,0xD7,0xF6,0x95,0xB4,
						0x5B,0x7A,0x19,0x38,0xDF,0xFE,0x9D,0xBC,
						0xC4,0xE5,0x86,0xA7,0x40,0x61,0x02,0x23,
						0xCC,0xED,0x8E,0xAF,0x48,0x69,0x0A,0x2B,
						0xF5,0xD4,0xB7,0x96,0x71,0x50,0x33,0x12,
						0xFD,0xDC,0xBF,0x9E,0x79,0x58,0x3B,0x1A,
						0xA6,0x87,0xE4,0xC5,0x22,0x03,0x60,0x41,
						0xAE,0x8F,0xEC,0xCD,0x2A,0x0B,0x68,0x49,
						0x97,0xB6,0xD5,0xF4,0x13,0x32,0x51,0x70,
						0x9F,0xBE,0xDD,0xFC,0x1B,0x3A,0x59,0x78,
						0x88,0xA9,0xCA,0xEB,0x0C,0x2D,0x4E,0x6F,
						0x80,0xA1,0xC2,0xE3,0x04,0x25,0x46,0x67,
						0xB9,0x98,0xFB,0xDA,0x3D,0x1C,0x7F,0x5E,
						0xB1,0x90,0xF3,0xD2,0x35,0x14,0x77,0x56,
						0xEA,0xCB,0xA8,0x89,0x6E,0x4F,0x2C,0x0D,
						0xE2,0xC3,0xA0,0x81,0x66,0x47,0x24,0x05,
						0xDB,0xFA,0x99,0xB8,0x5F,0x7E,0x1D,0x3C,
						0xD3,0xF2,0x91,0xB0,0x57,0x76,0x15,0x34,
						0x4C,0x6D,0x0E,0x2F,0xC8,0xE9,0x8A,0xAB,
						0x44,0x65,0x06,0x27,0xC0,0xE1,0x82,0xA3,
						0x7D,0x5C,0x3F,0x1E,0xF9,0xD8,0xBB,0x9A,
						0x75,0x54,0x37,0x16,0xF1,0xD0,0xB3,0x92,
						0x2E,0x0F,0x6C,0x4D,0xAA,0x8B,0xE8,0xC9,
						0x26,0x07,0x64,0x45,0xA2,0x83,0xE0,0xC1,
						0x1F,0x3E,0x5D,0x7C,0x9B,0xBA,0xD9,0xF8,
						0x17,0x36,0x55,0x74,0x93,0xB2,0xD1,0xF0 };

const char *monthn[]={"***","Jan","Feb","Mar","Apr","May","Jun",
									"Jul","Aug","Sep","Oct","Nov","Dec"};


void cleanlastname(char *n);
void cleanfirstname(char *n);
boolean ifprefix(char *b);
boolean ifprefix1(char *b);

#ifndef __OS2__

 void Strcpy(char huge *dest, char *source);
 void strcpY(char *dest, char huge *source);
 int Strcmp(char huge *s1, char huge *s2);

#endif

int cmp_nodenr(const void *elem1, const void *elem2);
int cmpnodenr(unsigned int *elem1, unsigned int *elem2);
boolean findnodelist(void);
int mergelists(char *rl, char *ol, int maxpoints);
char readline(FILE *inf, int linenr, unsigned long nr_records);
int getcommands(int argc, char *argv[]);
int scanoption(char *s, int maxl);
int readconfig(char *s, int maxl);
boolean substenvirons(char *s, int max);
int directory(char *match, char *fname);
boolean forbiddenext(char *f);
void findlist(char *fbuf, char *inlist, char *defname, char *defext,int ern);
void timestamp(FILE *f,int s);
char *Fgets(char *s,int n, FILE *fp);
void Error(int er, int ln, FILE *f, fpos_t pos, int cpos, boolean errab);
void help(void);
void pstring(char *st,int n);
void pvterror(int l,char *b);
unsigned int chcrc(unsigned char ch, unsigned int crc);
unsigned int Writeline(char *b, FILE *f, unsigned int ch);
int getchbeep(int maxtime);
boolean splitentry(char *b, char *st,char *nr,char *bsn,char *bl
												,char *bna,char *bnr,char *bd,char *bf);
void brackup(char *s, unsigned int size, byte type);
void remove_starC(char *us, int uf);
boolean isemptyline(char *s);

/************************** Main ********************/

int main(int argc, char *argv[])
{
	int h,i,j,k,l,m,n,np,r,nr_lines=0;
	unsigned int bosszone=Defaultzone,bossnet=DEFAULTREGION,bossnode=0;
	unsigned long freemem,nrofrecords,nrofbosses;
	char c,nam[NAMELENG],nummer[30],pointof[20],bossin[15];
	char ffdrv[MAXDRIVE], ffpath[MAXPATH], ffname[MAXFILE], ffext[MAXEXT];
	const char *monthn[]={"***","Jan","Feb","Mar","Apr","May","Jun",
									"Jul","Aug","Sep","Oct","Nov","Dec"};
	char huge *s;
	char huge *t;
	string version;
	longstring fbuf;
	struct date datum;
	fpos_t ccpos,bbpos;
	unsigned int checks;
	char yab[256],nodelistline[256];
	char status[15],bnumber[15];
	string bossysname, bossloc, bossname, bossnumber, bossbaud;
	longstring bossflags;
	int nlhandle, plhandle, pvthandle;
	FILE *nl;

	strcpy(version,VERSION);
	printf("\nPCMERGE %s, Pointlist/nodelist sorter/merger/convertor\n",version);
	printf(MYNAMEANDNUMBER);
	printf("\nNoncommercial use for Fidonet nodes and points allowed.\n");

	strcpy(spointlist,OUTPLISTN);		/* set some default filenames */
	strcat(spointlist,OUTPLISTE);
	strcpy(fdlist,FDFILEN);
	strcat(fdlist,FDFILEE);
	pointlist[0]=0;
	nodelist[0]=0;
	outlist[0]=0;
	logfile[0]=0;

	if (argc==2 && (argv[1][0]=='?' || toupper(argv[1][0])=='H'))
	{
		help();
		exit(0);
	}

	pvt_list=farcalloc(NRPRIVATS,sizeof(longstring));
	if (pvt_list==NULL) Error(99,0,0,0,0,TRUE);
	pvtlist=pvt_list;
	for (i=0;i<NRPRIVATS;i++) Strcpy(pvtlist[i],"");

	if ((i=readconfig(yab,sizeof(yab)-1))!=0)
	{
		printf("\nError reading config file line %d\n",abs(i));
		if (i>0)	printf("unknown option : '%s'\n",yab);
		else printf("unknown environment variable : '%s'\n",yab);
		exit(23);
	}

	if (argc>1)
	{
		if ((i=getcommands(argc, argv))!=0)
		{
		if (i>0)	printf("\nUnknown option : '%s'\n",argv[abs(i)]);
		else printf("\nUnknown environment variable : '%s'\n",argv[abs(i)]);
		exit(23);
		}
	}
	if (logflag)
	{
		if (logflag==2) log=fopen(logfile,"wt"); else log=fopen(logfile,"at");
		if (log==NULL) Error(4,0,0,0,0,TRUE);
		else
		{
         getdate(&datum); i=datum.da_mon;
         if (i<1 || i>12) i=0;
			fprintf(log,"\n----------  %d-%s-%d, PCMERGE v %s\n"
						,datum.da_day,monthn[i],datum.da_year,version);
		}
	}
	if (pwarning) {putch('\a');Error(2,0,0,0,0,FALSE);} /* too many private lists */
	freemem=farcoreleft()-5000;
	nrofrecords=freemem/(sizeof(p_entry)+(AVERNAMELEN+1));
	if (mergeflag || usebossname || usebossloc || usebossfone || excludeorphan)
		nrofrecords=nrofrecords-(nrofrecords*sizeof(fpos_t)/((sizeof(p_entry)+(AVERNAMELEN+1))*AVERPOINTS));
	if (nrofrecords<MINENTRYS) Error(99,0,0,0,0,TRUE);
	if (nrofrecords>MAXENTRYS)
	{
		nrofrecords=MAXENTRYS;
		printf("\nPlenty of m");
	}
	else printf("\nM");
	printf("emory for about %u points",nrofrecords);
	p_l1=farcalloc(nrofrecords,sizeof(p_entry));
	p_name=farcalloc(nrofrecords*(AVERNAMELEN+1),sizeof(char));
	if (p_l1==NULL || p_name==NULL) Error(99,0,0,0,0,TRUE);
	p_list=p_l1;
	pname=last_pname=p_name;

	if (mergeflag || usebossname || usebossloc || usebossfone || excludeorphan)
	{
		freemem=farcoreleft()-5000;
		nrofbosses=freemem/sizeof(fpos_t);

		bosspos1=farcalloc(nrofbosses,sizeof(fpos_t));
		if (bosspos1==NULL)  {printf("\n");Error(99,0,0,0,0,TRUE);}
		bosspos=bosspos1;
		aantbosses=(int)(nrofbosses & 0x7FFF);
		printf(" and %u bosses",nrofbosses);
	}
	printf(".\n");

	/*************  read the input list into memory ************/

	findlist(fbuf,pointlist,INPLISTN,".???",16);
	plhandle=open(fbuf,O_RDONLY|O_DENYNONE);
	if (plhandle==-1) Error(15,0,0,0,0,TRUE);
	infile=fdopen(plhandle,"rt");
	if (infile==NULL) Error(15,0,0,0,0,TRUE);
	fnsplit(fbuf,NULL,NULL,NULL,pext);
	sprintf(pstr,"Reading pointlist: %s\n",fbuf);
	pstring(pstr,0);
	while ((r=readline(infile,nr_lines,nrofrecords ))!=EOF)
	{
		if (r==0) continue;			/* skip line in error */
		nr_lines++;
		if (nr_lines>=nrofrecords)
		{
			putch('\a');
			sprintf(pstr,"Warning: Not enough memory, not all points read!\n");
			pstring(pstr,2);
			if (errornumber==0) errornumber=1;
			break;
		}
	}
	fclose(infile);
	sprintf(pstr,"%d Points read.\n",nr_lines);
	pstring(pstr,0);

	/****************** Check for dup names in same net ************/

	if (npcflag)
	{
		for (k=0;k<nr_lines-1;k++)
		{
			if (npcflag==TRUE) h=nr_lines; else h=k+2;
			for (i=k+1;i<h;i++)
			{
				s=p_list[index[k]].namep;
				t=p_list[index[i]].namep;

				if (Strcmp(s,t)==0)
				if (p_list[index[k]].zone==p_list[index[i]].zone
										&& p_list[index[k]].net==p_list[index[i]].net)
				{
					nrofdups++;
					if (nrofdups==1)
					{
						sprintf(pstr,"Duplicate names found in same net:\n");
						pstring(pstr,2);
						sprintf(pstr,"Line nrs.  Name                 Nodenumbers\n");
						pstring(pstr,0);
					}
					for (j=0;j<sizeof(pstr);j++) pstr[j]=' ';
					for (j=0;j<NAMELENG;j++)
					{
						c=(s)[j];
						if (c!='_') nam[j]=c; else nam[j]=' ';
						if (c==0) break;
					}
					if (j<20) for (;j<20;j++) nam[j]=' ';
					nam[j]=0;
					j=sprintf(pstr,"%4d-%4d  %s",k,i,nam);
					j=j+sprintf(&pstr[j]," %u:%u/%u.%u",p_list[index[k]].zone,p_list[index[k]].net,
														p_list[index[k]].node,p_list[index[k]].point);
					if (j<50) for(;j<50;j++) pstr[j]=' ';
					pstr[j]=0;
					sprintf(&pstr[j]," %u:%u/%u.%u\n",p_list[index[i]].zone,p_list[index[i]].net,
														p_list[index[i]].node,p_list[index[i]].point);
					pstring(pstr,0);
				}
			}
		}
		if (nrofdups>0)
		{
			if (errornumber==0) errornumber=1;
			sprintf(pstr,"%d duplicate names found.\n",nrofdups);
			pstring(pstr,0);
		}
	}

	/**********************  Sort by nodenumber  **************/

	sprintf(pstr,"Sorting points.\n");
	pstring(pstr,0);
	qsort(index,nr_lines,sizeof(unsigned int),cmp_nodenr);

	/******* remove numbers that are also found in the private list *****/

	for (n=0;n<=privats;n++)
	{
		np=0;
		if (n==0) strcpy(nam,PVTLISTN); else nam[0]=0;
		strcpY(pstr,pvtlist[n]);
		findlist(fbuf,pstr,nam,".???",21);
		pvthandle=open(fbuf,O_RDONLY|O_DENYNONE);
		if (pvthandle==-1) Error(22,0,0,0,0,TRUE);
		infile=fdopen(pvthandle,"rt");
		if (infile==NULL) Error(22,0,0,0,0,TRUE);
		sprintf(pstr,"Private pointlist: %s\n",fbuf);
		pstring(pstr,0);
		bosszone=Defaultzone;
		bossnet=DEFAULTREGION;
		bossnode=-1;

		l=0;
		while (Fgets(fbuf,sizeof(fbuf),infile)!=NULL)
		{
			l++;
			if (fbuf[0]==';' || isemptyline(fbuf)) continue;
			for (j=0;j<7;j++) nam[j]=toupper(fbuf[j]);
			if (strncmp(nam,"BOSS",4)==0)
			{
				j=sscanf(&fbuf[5],"%d:%d/%d",&bosszone,&bossnet,&bossnode);
				if (j!=3)
				{
					pvterror(l,fbuf);
					bossnode=-1;
				}
				continue;
			}
			m=1;
			if (strncmp(nam,"PVT",3)==0) m=4;
			if (strncmp(nam,"HOLD",4)==0) m=5;
			if (strncmp(nam,"DOWN",4)==0) m=5;

			j=sscanf(&fbuf[m],"%d",&k);
			if (j!=1 || fbuf[m-1]!=',')
			{
				pvterror(l,fbuf);
				continue;
			}
			for (i=0;i<nr_lines;i++)
			{
				if (	((p_list[index[i]].zone == bosszone) &&
						(p_list[index[i]].net  == bossnet)  &&
						(p_list[index[i]].node == bossnode) &&
						(p_list[index[i]].point== k) ) )  /* same number? */
				{
					np++;
					if ((p_list[index[i]].flags & 0x01)==0)
					{
						p_list[index[i]].flags |= 0x01;		/* mark as deleted */
						nrofprivs++;                  /* if not already marked */
					}
				}
			}
		}
		fclose(infile);
		strcpy(fbuf," private points to skip.");
		sprintf(pstr,"%d%s\n",np,fbuf);
		pstring(pstr,0);
		bosszone=Defaultzone;
		bossnet=DEFAULTREGION;
		bossnode=-1;
	}
	if (privats>0)
	{
		sprintf(pstr,"%d private points total in %d lists.\n"
															,nrofprivs,privats+1);
		pstring(pstr,0);
	}

	/************** check / remove duplicate nodenumbers ********/

	if (multiflag != 0)	/* check/remove duplicates if enabled */
	{
		nrofdups=0;
		for (i=1;i<nr_lines;i++)
		{
			if (	(p_list[index[i]].zone == p_list[index[i-1]].zone) &&
					(p_list[index[i]].net  == p_list[index[i-1]].net)  &&
					(p_list[index[i]].node == p_list[index[i-1]].node) &&
					(p_list[index[i]].point== p_list[index[i-1]].point) )  /* same number? */
			{
				if (errornumber==0) errornumber=1;
				nrofdups++;
				if (nrofdups==1)
				{
					sprintf(pstr,"Multiple entrys found:\n");
					pstring(pstr,2);
					strcpy(pstr,"Line nr.   Node number       Name   ");
					if (multiflag==2) strcat(pstr,"removed");
											else strcat(pstr,"       ");
					strcat(pstr,"       Name ");
					if (multiflag==2) strcat(pstr,"retained");
					strcat(pstr,"\n");
					pstring(pstr,0);
				}
				k=sprintf(pstr,"%4d-%4d  %u:%u/%u.%u ",index[i-1]+1,index[i]+1
								,p_list[index[i-1]].zone,p_list[index[i-1]].net
								,p_list[index[i-1]].node,p_list[index[i-1]].point);
				if (k<=28) for (;k<=28;k++) pstr[k]=' ';
				pstr[k]=0;
				j=0;
				while (TRUE)
				{
					s=p_list[index[i-1]].namep;
					c=(s)[j];
					if (c==0) break;
					if (c=='_') c=' ';
					nam[j++]=c;
				}
				nam[j]=0;
				strcat(pstr,nam);
				if ((j=strlen(pstr))<50) for (;j<50;j++) pstr[j]=' ';
				pstr[j]=0;
				j=0;
				while (TRUE)
				{
					s=p_list[index[i]].namep;
					c=(s)[j];
					if (c==0) break;
					if (c=='_') c=' ';
					nam[j++]=c;
				}
				nam[j]=0;
				strcat(pstr,nam);
				strcat(pstr,"\n");
				pstring(pstr,0);
				if (multiflag==2) p_list[index[i-1]].flags|=0x01; /* mark deleted */
			}
		}
	}
	if ((multiflag !=0) && (nrofdups !=0))
	{
		sprintf(pstr,"%d duplicate nodenumbers ",nrofdups);
		if (multiflag==1) strcat(pstr,"found.\n"); else strcat(pstr,"removed.\n");
		pstring(pstr,0);
	}

	/************** print human readable list sorted by number ********/

	if (sortflag)
	{
      if ((fnsplit(spointlist,ffdrv,ffpath,ffname,ffext) & FILENAME) ==0)
           fnmerge(spointlist,ffdrv,ffpath,OUTPLISTN,OUTPLISTE);
      outfile=fopen(spointlist,"wt");
		if (outfile==NULL) Error(10,0,0,0,0,TRUE);
		sprintf(pstr,"Writing sorted pointlist: %s\n",spointlist);
		pstring(pstr,0);
		for (i=0;i<nr_lines;i++)
		{
			if (p_list[i].flags & 0x01) continue;		/* skip deleted entry */
			j=0;
			while (TRUE)
			{
				s=p_list[index[i]].namep;
				c=(s)[j];
				if (c==0) break;
				if (c=='_') c=' ';
				nam[j++]=c;
			}
			nam[j]=0;
			sprintf(nummer,"%u:%u/%u.%u",p_list[index[i]].zone,p_list[index[i]].net,
			p_list[index[i]].node,p_list[index[i]].point);
			if (strlen(nummer)<18)
			for (j=strlen(nummer);j<17;j++) strcat(nummer," ");
			fprintf(outfile,"%s   %s\n",nummer,nam);
		}
		fclose(outfile);
	}

	/*************** Merge pointlist with region segment ************/

	if (mergeflag || usebossname || usebossloc || usebossfone || excludeorphan)
	{
		if (findnodelist() && npcflag)
		{
			sprintf(pstr,"Warning: %s is older than %d days\n",nodelist,MAXDAYS);
			pstring(pstr,1);
			printf("Continue? (Y/N) ");
			i=toupper(getchbeep(300));
			if (i!='Y' && i!='J' && i!='\r') Error(98,0,0,0,0,TRUE);
			printf("\n");
		}
		if (mergeflag)
		{
			sprintf(pstr,"Merging pointlist with nodelist: %s\n",nodelist);
			pstring(pstr,0);
			sprintf(pstr,"Output written to : %s\n",outlist);
			pstring(pstr,0);
		}
		else
		{
			sprintf(pstr,"Reading nodelist: %s\n",nodelist);
			pstring(pstr,0);
		}
		j=mergelists(nodelist,outlist,nr_lines);
		sprintf(pstr,"%u Nodes read, %d of them have points.\n",an,j);
		pstring(pstr,0);
		if (mergeflag)
		{
			sprintf(pstr,"%d Points merged.\n",ap);
			pstring(pstr,0);
		}
		if (multiflag==2) k=nr_lines-ap-nrofdups-nrofprivs;
											else k=nr_lines-ap-nrofprivs;
		if (k)  /* report points without a boss in te nodelist(segment) */
		{
			if (errornumber==0) errornumber=1;
			sprintf(pstr,"%d Orphaned points :\n",k);
			pstring(pstr,2);
			sprintf(pstr,"Line  Nodenumber         Name\n");
			pstring(pstr,0);
			for (i=0;i<nr_lines;i++)
			{
				if ((p_list[index[i]].flags & 0x81)==0)
				{
					k=sprintf(pstr,"%4d  %u:%u/%u.%u",index[i]+1,
						p_list[index[i]].zone,p_list[index[i]].net,
						p_list[index[i]].node,p_list[index[i]].point);
					j=k;
					if (k<25) for (j=k;j<25;j++) pstr[j]=' ';
					pstr[j]=0;
					j=0;
					while (TRUE)
					{
						s=p_list[index[i]].namep;
						c=(s)[j];
						if (c==0) break;
						if (c=='_') c=' ';
						nam[j++]=c;
					}
					nam[j]=0;
					strcat(pstr,nam);
					strcat(pstr,"\n");
					pstring(pstr,0);
				}
			}
		}
	}

	/*****************  FDPOINT.PVT style St.Louis list ***************/

	if (fdflag)
	{
		if (usebossname || usebossloc || usebossfone)
		{
			nlhandle=open(nodelist,O_RDONLY|O_DENYNONE);
			if (nlhandle==-1)  Error(14,0,0,0,0,TRUE);
			nl=fdopen(nlhandle,"rt");
			if (nl==NULL)  Error(14,0,0,0,0,TRUE);
		}
		if ((fnsplit(fdlist,ffdrv,ffpath,ffname,ffext) & FILENAME) ==0)
			  fnmerge(fdlist,ffdrv,ffpath,FDFILEN,FDFILEE);
      outfile=fopen(fdlist,"wt");
		if (outfile==NULL) Error(10,0,0,0,0,TRUE);
		sprintf(pstr,"Writing: %s\n",fdlist);
		pstring(pstr,0);
		getdate(&datum);
		i=datum.da_mon; if (i<1 || i>12) i=0;
		if (isdigit(pext[1]) && isdigit(pext[2]) && isdigit(pext[3]) )
		{
			strcpy(nam,&pext[1]);
		}
		else strcpy(nam,"000");
		fprintf(outfile,";S Pointlist created by PCMERGE %s on %d-%s-%d -- number %s : "
							,version,datum.da_day,monthn[i],datum.da_year,nam);
		fgetpos(outfile,&ccpos);
		fprintf(outfile,"00000\n");
		checks=0;
		if (usebossname || usebossloc || usebossfone)
		{
			sprintf(pstr,";S Names from: %s\n",nodelist);
			pstring(&pstr[3],0);
			checks=Writeline(pstr,outfile,checks);
		}
		ap=k=0;
		for (i=0;i<nr_lines;i++)
		{
			if (p_list[index[i]].flags & 0x01) continue;		/* skip deleted entry */
			if (excludeorphan && (p_list[index[i]].flags & 0x80)==0) continue; /* skip orphans */
			j=0;
			while (TRUE)
			{
				s=p_list[index[i]].namep;
				c=(s)[j];
				if (c==0) break;
				nam[j++]=c;
			}
			nam[j]=0;
			if (bosszone!=p_list[index[i]].zone||bossnet!=p_list[index[i]].net||
				 bossnode!=p_list[index[i]].node)
			{
				bosszone=p_list[index[i]].zone;
				bossnet=p_list[index[i]].net;
				bossnode=p_list[index[i]].node;
				sprintf(yab,"Boss,%u:%u/%u,\n",bosszone,bossnet,bossnode);
				checks=Writeline(yab,outfile,checks);
				if (usebossname || usebossloc || usebossfone)
				{
					if (p_list[index[i]].bossnr>=0)
					{
						bbpos=bosspos[p_list[index[i]].bossnr];
						if (fsetpos(nl,&bbpos)) Error(14,0,0,0,0,TRUE);
						Fgets(nodelistline,sizeof(nodelistline)-1,nl);
						if (!splitentry(nodelistline,status,bnumber,bossysname,
								bossloc,bossname,bossnumber,bossbaud,bossflags))
															Error(14,0,0,0,0,TRUE);
					}
					else
					{
						strcpy(bossname,"-Error-");
						strcpy(bossloc,"-Error-");
						strcpy(bossysname,"-Error-");
						strcpy(bossnumber,"-Unpublished-");
						strcpy(bossbaud,"2400");
						bossflags[0]=0;
					}
				}
				k++;
			}
			pointof[0]=0;
			if (usebossname)
			{
				if (p_list[index[i]].flags & 0x80)
				{
					if (usebossname & 0x08)	strcpy(sysname,bossname);
					else strcpy(sysname,bossysname);
				}
				else strcpy(sysname,"-Boss_unlisted-");
				if (usebossname & 0x4)
				{
					sprintf(pointof,"Point_%u_of_",p_list[index[i]].point);
				}
				if (usebossname & 0x02)
				{
					brackup(sysname,sizeof(sysname),2);
					if (usebossname & 0x10) brackup(pointof,sizeof(pointof),1);
					else brackup(sysname,sizeof(sysname),1);
				}
			}
			if (usepointname) strcpY(sysname,p_list[index[i]].namep);

			bossin[0]=0;
			if (usebossloc)
			{
				if (p_list[index[i]].flags & 0x80)
				{
					strcpy(location,bossloc);
				}
				else strcpy(location,"-Unknown-");
				if (usebossloc & 0x04) strcpy(bossin,"Boss_in_");
				if (usebossloc & 0x02)
				{
					brackup(location,sizeof(location),2);
					if (usebossloc & 0x10) brackup(bossin,sizeof(bossin),1);
					else brackup(location,sizeof(location),1);
				}
			}

			if (usebossfone)
			{
				if (p_list[index[i]].flags & 0x80)
				{
					if (!(usebossfone & 0x02))
					{
						strcpy(pointphonenr,bossnumber);
					}
					strcpy(baud,bossbaud);
					strcat(baud,",");
					strcat(baud,bossflags);
				}
				else
				{
					if (!(usebossfone & 0x02))
					{
						strcpy(pointphonenr,"-Unpublished-");
						strcpy(baud,"2400,");
					}
				}
			}
			sprintf(yab,",%u,%s%s,%s%s,%s,%s,%s\n", p_list[index[i]].point, pointof,
				 sysname, bossin, location, p_list[index[i]].namep, pointphonenr, baud);
			checks=Writeline(yab,outfile,checks);
			ap++;
		}
		if (usebossname || usebossloc || usebossfone) fclose(nl);
		fsetpos(outfile,&ccpos);
		fprintf(outfile,"%05u",checks);
		fclose(outfile);
				sprintf(pstr,"Checksum : %05u\n",checks);
		pstring(pstr,0);
		sprintf(pstr,"%d points and %d bosses written.\n",ap,k);
		pstring(pstr,0);
	}

	/********************   All done, clean up and exit  **************/

	if (zapflag && (errornumber==0))
	{
		if (remove(pointlist)) strcpy(fbuf,"could not be deleted.");
								else  strcpy(fbuf,"succesfully deleted");
		sprintf(pstr,"%s %s\n",pointlist,fbuf);
		pstring(pstr,2);
	}
	printf("Ready...\n\n");
	if (usebossname || usebossloc || usebossfone) farfree(bosspos1);
	farfree(p_name);
	farfree(p_l1);
	if (logflag)
	{
		timestamp(log,0);
		fprintf(log,"PCMERGE completed.\n");
		fclose(log);
	}
	farfree(pvt_list);
	return(errornumber);
}

boolean findnodelist(void)
{
	char f_drv[MAXDRIVE], f_path[MAXPATH], f_name[MAXFILE], f_ext[MAXEXT];
	char ff_drv[MAXDRIVE], ff_path[MAXPATH], ff_name[MAXFILE], ff_ext[MAXEXT];
	longstring f_buf;
	int i;

   #if defined (__WATCOMC__) && defined (__OS2__)

     struct stat mystat;

   #else

     struct ftime ftimep;
     struct date fdate;
     struct time ftime;

   #endif

	time_t timer;
	long filetime,fileage,days;
	int fhandle;


	findlist(f_buf,nodelist,REGFILE,".???",17);
	i=fnsplit(nodelist,ff_drv,ff_path,ff_name,ff_ext);
	if (outlist[0]==0)
	{
		fnmerge(outlist,ff_drv,ff_path,ff_name,".999");
	}
	else
	{
		i=fnsplit(outlist,f_drv,f_path,f_name,f_ext);
		if ((i & FILENAME) == 0) strcpy(f_name,ff_name);
		if ((i & EXTENSION) == 0) strcpy(f_ext,".999");
		fnmerge(outlist,f_drv,f_path,f_name,f_ext);
	}

	fhandle=open(nodelist,O_RDONLY|O_DENYNONE);
   if (fhandle==-1) Error(14,0,0,0,0,TRUE);

   #if defined (__WATCOMC__) && defined (__OS2__)

     if(fstat(fhandle, &mystat) == -1) {close(fhandle); Error(14,0,0,0,0,TRUE);}
     close(fhandle);
     filetime=mystat.st_mtime;

   #else

     if (getftime(fhandle,&ftimep)==-1) {close(fhandle); Error(14,0,0,0,0,TRUE);}
     close(fhandle);
     fdate.da_year=ftimep.ft_year+1980;
     fdate.da_mon=ftimep.ft_month;
     fdate.da_day=ftimep.ft_day;
     ftime.ti_hour=ftimep.ft_hour;
     ftime.ti_min=ftimep.ft_min;
     ftime.ti_sec=ftimep.ft_tsec;
     ftime.ti_hund=0;
     filetime=dostounix(&fdate,&ftime);     /* filetime naar seconden */

   #endif

	fileage=time(&timer)-filetime;
	days=(fileage+42200l)/(86400l);			/* seconden naar dagen */
	return(days>MAXDAYS);
}

int getchbeep(int maxtime)
{

	time_t secsnow, secs, secs1;

	secs=secs1=time(&secsnow)-15;
	while (!kbhit())
	{
		time(&secsnow);
		if ((secsnow-secs1)>maxtime) return(0);
		if (secsnow-secs>15)
		{
         #ifdef __OS2__
          DosBeep(600, 300);
         #else
          sound(600);
          delay(300);
         #endif

			if (kbhit()) break;

         #ifdef __OS2__
          DosBeep(800, 300);
         #else
          sound(800);
          delay(300);
         #endif

			if (kbhit()) break;

         #ifdef __OS2__
          DosBeep(1000, 300);
         #else
          sound(1000);
          delay(300);
			 nosound();
         #endif

			secs=secsnow;
		}
	}

   #ifndef __OS2__
      nosound();
   #endif

	return(getche());
}


#ifndef __OS2__

/* copy to a string on the far heap with a huge pointer */

void Strcpy(char huge *dest, char *source)
{
	int i=0;

	while (source[i] && i<(sizeof(string)-1)) (dest)[i]=source[i++];
	(dest)[i]=0;
}

/* copy from a string on the far heap with a huge pointer */

void strcpY(char *dest, char huge *source)
{
	int i=0;

	while ((source)[i] && i<(sizeof(string)-1)) dest[i]=(source)[i++];
	dest[i]=0;
}

/* compare two strings on the far heap with huge pointers */

int Strcmp(char huge *s1, char huge *s2)
{
	int i=0;

	while ((s1)[i] && (s2)[i] && (s1)[i]==(s2)[i]) i++;
	return((s1)[i]-(s2)[i]);
}

#endif

/* Compare nodenumbers for sort */

int cmp_nodenr(const void *elem1, const void *elem2)
{
	return(cmpnodenr((unsigned int*)elem1,(unsigned int*)elem2));
}

int cmpnodenr(unsigned int *elem1, unsigned int *elem2)
{
	int r;
	p_entry p,q;

	p=p_list[*elem1];
	q=p_list[*elem2];

	r=p.zone - q.zone;
	if (r==0)
	{
		r=p.net - q.net;
		if (r==0)
		{
			r=p.node - q.node;
			if (r==0)
			{
				r=p.point - q.point;
			}
		}
	}
	if (r>0) r=1;
	if (r<0) r=-1;
	return(r);
}

/* Read one line from input pointlist */

char readline(FILE *inf, int linenr, unsigned long nr_records)
{
	char b=0,c=0,d=0;
	char fname[80],lname[40];
	int i=0, j=0, k=0, n=0, temp=0, col=0, colm=0, colum=0;
	boolean commafound=FALSE, wrongc=FALSE, wrongchar=FALSE, zonefound=FALSE;
	fpos_t pos, pos1;

	fgetpos(inf, &pos);		/* remember position in file in case of error */

	while	((c!=EOF && c!=',') && i<40-1)
	{
		c=fgetc(inf);
		if (c==0x1A && k==0) return(EOF);
		if (c=='\n') {Error(6,linenr+1,inf,pos,n+2,FALSE); return(0);}
      if (c==';' && n==0)              /* comment line ? */
		{
			while (c!=EOF && c!='\n') c=fgetc(inf);	/* skip comment line */
			fgetpos(inf, &pos);								/* update last pos */
			k=0;
			continue;
		}
		k++;	n++;
		if (npcflag)
		{
			if ( !isprint(c) || ( (c=='.' || c=='_' || c==' ') && c!=separator))
			{
				if (!wrongchar) { wrongchar=TRUE; colum=n; }
			}
		}
		switch (c)
		{
			case '_'	:
			case ' '	:  d='_';	break;
			case '.'	:	if (nameclean>=0) d='_'; else d=c; break;
			case ','	:	d=0; commafound=TRUE; break;
			default	:	d=c;		break;
		}
		lname[i++]=d;
	}
	lname[i]=0;
	for (j=i-1;j>0;j--) { if (lname[j]=='_') lname[j]=0; else break; }
	if ((!commafound) && colum==0) colum=n-(i-j);
	if (nameclean==1) cleanlastname(lname);

	if (i>=40-1) while (c!=','&&c!=' '&&c!='\n'&&c!=EOF) { c=fgetc(inf); n++; }
	if (c==',') commafound= TRUE;
	if (c=='\n') {Error(6,linenr+1,inf,pos,n+2,FALSE); return(0);}
	index[linenr]=linenr;

	fgetc(inf); n++;
	if (c=='\n') {Error(6,linenr+1,inf,pos,n,FALSE); return(0);}
	c=0;
	i=0;
	if (commafound)
	{
		while	((c!=EOF && (c!=' ' || b!=' ')) && i<40-1)
		{
			b=c;
			c=fgetc(inf);
			if (c=='\n') {Error(6,linenr+1,inf,pos,n,FALSE); return(0);}
			n++;
			if (npcflag)
			{
				if ( !isprint(c) || ( (c=='.' || c=='_' || c==' ') && c!=separator))
				{
					if (!wrongc) { wrongc=TRUE; col=n; }
				}
			}
			switch (c)
			{
				case '_'	:
				case ' ' :  d='_';	break;
				case '.'	:	if (nameclean>=0) d='_'; else d=c; break;
				case '\n':	Error(6,linenr+1,inf,pos,n-1,FALSE);
								return(0);
				default	:	d=c;		break;
			}
			fname[i++]=d;
		}
		fname[i]=0;

		if (wrongc && col<=n-2 && !wrongchar) { wrongchar=TRUE; colum=col; }
		if (i>1 && fname[i-1]=='_' && fname[i-2]=='_') fname[i-1]=0;
		if (strlen(fname)+strlen(lname)>NAMELENG-1)
		{
			temp=strlen(fname)+strlen(lname)-NAMELENG+1;
			if (temp>strlen(lname)) {fname[NAMELENG-1]=0; lname[0]=0;}
			else lname[NAMELENG-strlen(fname)-1]=0;	/*truncate if too long */
		}
		if (nameclean==1) cleanfirstname(fname);
		if (fname[0]=='_' && fname[1]==0)     /* missing first name */
		{
			fname[0]=0;
			c=' ';
			fgetpos(inf,&pos1);
			Error(3,linenr+1,inf,pos,n-1,FALSE);
			fsetpos(inf,&pos1);
		}
		strcat(fname,lname);
		fname[NAMELENG-1]=0;		/* truncate */
	}
	else
	{
		strcpy(fname,lname);
		fname[NAMELENG-1]=0;		/* truncate */
		c=' ';
	}
	totalnamechars=totalnamechars+strlen(fname);
	p_list[index[linenr]].bossnr=-1;
	p_list[index[linenr]].namep=last_pname;
	Strcpy(last_pname,fname);
	last_pname=last_pname+strlen(fname)+1;
	if (last_pname+NAMELENG>pname+(nr_records*(AVERNAMELEN+1))) Error(99,linenr+1,0,0,0,TRUE);
	if (i>=40-1) while (c!=' ' || c!='\t')
	{
		c=fgetc(inf);
		if (c=='\n') {Error(6,linenr+1,inf,pos,n-1,FALSE); return(0);}
		n++;
	}
	while ((c==' ' || c=='\t') && c!=EOF)
	{
		c=fgetc(inf);
		if (c=='\n') {Error(6,linenr+1,inf,pos,n-1,FALSE); return(0);}
		n++;
	}
	temp=0;
	colm=n-1;
	while (isdigit(c))
	{
		temp=temp*10;temp+=(c-'0');
		c=fgetc(inf);
      if (c=='\n') {Error(6,linenr+1,inf,pos,n-1,FALSE); return(0);}
		n++;
	}
	if (c==EOF) return(c);
	if (c!='/' && c!=':') { Error(5,linenr+1,inf,pos,n,FALSE); return(0); }
	if (c=='/')  p_list[linenr].net=temp;
	if (c==':')
	{
		zonefound=TRUE;
		p_list[linenr].zone=temp;
		temp=0;
		c=fgetc(inf);
      if (c=='\n') {Error(6,linenr+1,inf,pos,n-1,FALSE); return(0);}
		n++;
		while (isdigit(c))
		{
			temp=temp*10; temp+=(c-'0');
			c=fgetc(inf);
			n++;
		}
		if (c==EOF) return(c);
		p_list[linenr].net=temp;
		if (c!='/') { Error(5,linenr+1,inf,pos,n,FALSE); return(0); }
	}
	else
	{
		if (colum==0) colum=colm;
		p_list[linenr].zone=Defaultzone;
		p_list[linenr].net=temp;
	}

	temp=0;
	c=fgetc(inf);
	n++;
	while (isdigit(c))
	{
		temp=temp*10; temp+=(c-'0');
		c=fgetc(inf);
		n++;
	}
	if (c==EOF) return(c);
	if (c!='.') { Error(5,linenr+1,inf,pos,n,FALSE); return(0); }
	p_list[linenr].node=temp;

	temp=0;
	c=fgetc(inf);
	n++;
	while (isdigit(c)) /* && c!=EOF) */
	{
		temp=temp*10; temp+=c-'0';
		c=fgetc(inf);
		n++;
	}
	p_list[linenr].point=temp;
	if (temp==0) { Error(5,linenr+1,inf,pos,n-1,FALSE); return(0); }
	if (c=='\n' || (c==EOF &&k==0)) return(c);
	while (c!='\n' && c!=EOF) c=fgetc(inf);
	if (npcflag)
	{
		if (!commafound || wrongchar) Error(1,linenr+1,inf,pos,colum,FALSE);
		if (!zonefound) Error(0,linenr+1,inf,pos,colum,FALSE);
	}
	if (c==EOF && k) return(1); else return(c);
}

/* Clean up last name, by substituting spaces for dec. points
	and properly capitalizing van's, der's etc.  */

void cleanlastname(char *n)
{
	char buffer[40];
	int i,j,k;

	if (n[0]==0) return;

	i=k=0;
	while (TRUE)
	{
		j=0;
		k=i;
		buffer[0]=0;
		while (n[i] !='_' && n[i] != 0) buffer[j++]=n[i++];
		buffer[j++]=0;
		if (buffer[0]==0) break;
		if (ifprefix(buffer)) n[k]=tolower(n[k]);
		else
		{
			n[k]=toupper(n[k]);
			if (toupper(buffer[0])=='T' && buffer[1]=='\'' && buffer[2])
			{
				n[k]=tolower(n[k]);
				n[k+2]=toupper(n[k+2]);
			}
		}
		if (n[i]==0) break; else i++;
	}
}

/* check if a word is a name prefix */

boolean ifprefix(char *b)
{
	if	(stricmp(b,"VAN")==0 || stricmp(b,"TER")==0 ||
		 stricmp(b,"DE") ==0 || stricmp(b,"IN")==0  ||
		 stricmp(b,"DEN")==0 || stricmp(b,"TE")==0  ||
		 stricmp(b,"LE")==0  || stricmp(b,"TEN")==0 ||
		 stricmp(b,"VD")==0  || stricmp(b,"DU")==0  ||
		 stricmp(b,"V")==0   || stricmp(b,"DER")==0 ||
		 stricmp(b,"D")==0   || stricmp(b,"OP")==0  ||
		 stricmp(b,"HET")==0 || stricmp(b,"VON")==0 ||
		 stricmp(b,"LA")==0  || stricmp(b,"T")==0 )
		 return(TRUE);
	else
		 return(FALSE);
}

boolean ifprefix1(char *b)
{
	if	(stricmp(b,"VAN")==0 || stricmp(b,"TER")==0 ||
		 stricmp(b,"DE") ==0 || stricmp(b,"IN")==0  ||
		 stricmp(b,"DEN")==0 || stricmp(b,"TE")==0  ||
		 stricmp(b,"LE")==0  || stricmp(b,"TEN")==0 ||
		 stricmp(b,"VD")==0  || stricmp(b,"DU")==0  ||
		 stricmp(b,"V")==0   || stricmp(b,"DER")==0 ||
		 stricmp(b,"OP")==0  || stricmp(b,"LA")==0  ||
		 stricmp(b,"HET")==0 || stricmp(b,"VON")==0 )
		 return(TRUE);
	else
		 return(FALSE);
}

/* clean up first name */

void cleanfirstname(char *n)
{
	char buffer[40];
	int i=1,j,k;

	if (n[0]==0) return; else n[0]=toupper(n[0]);

	while(n[i])
	{
		if (n[i-1]=='-') { n[i]=toupper(n[i]); i++; continue; }
		if (n[i-1]=='_')
		{
			n[i]=toupper(n[i]);
			if (i==2 && isalpha(n[0])) { n[i-1]='.'; i++; continue; }
			if (i>2 && isalpha(n[i-2]) && !isalpha(n[i-3]))
			{
				n[i-1]='.'; i++; continue;
			}
			if (n[i]=='_') n[i-1]='.';
			j=0; k=i;
			while (n[k] !='_' && n[k] != 0) buffer[j++]=n[k++];
			buffer[j++]=0;
			if (ifprefix1(buffer)) n[i]=tolower(n[i]);
		}
		i++;
	}
}

/* Merge pointlists with nodelist */

int mergelists(char *rl, char *ol, int maxpoints)

/* merge the nodelistsegment in rl with the point array in struct
	p_list. The name of the output file is in ol. */
{
	char buffer[256],status[15],bnumber[15],bossin[15];
	string bossysname, bossloc, bossname, bossnumber, bossbaud;
	longstring bossflags;
	int nextboss,nrbosses,i,m,n,z,p;
	int zonenr=Defaultzone,netnr=0,nodenr=0;
	int fhandle;
	byte bracktype;
	boolean bossfound,zoneflag,hostflag,firstline=TRUE,checkfound=FALSE;
	boolean toomanybosses=FALSE;
	unsigned int nmbr,crc;
	fpos_t cpos,bpos;
	struct date datum;

	crc=0;
	fhandle=open(rl,O_RDONLY|O_DENYNONE);
	if (fhandle==-1) Error(12,0,0,0,0,TRUE);
	infile=fdopen(fhandle,"rt");
	if (infile==NULL) Error(12,0,0,0,0,TRUE);

	if (mergeflag)
	{
		outfile=fopen(ol,"wt");
		if (outfile==0) Error(13,0,0,0,0,TRUE);
	}
	i=n=m=p=z=0;                           /* everything points to start */
	nrbosses=nextboss=an=ap=0;             /* number of nodes/points prcsd */
	while (TRUE)                           /* loop for all lines until EOF */
	{
		fgetpos(infile,&bpos);              /* get the pos of the boss line */
		if (fgets(buffer,sizeof(buffer)-1,infile)==NULL)	/* get a line from the infile */
		{                                   /* exit if EOF */
			fclose(infile);
			if (mergeflag)
			{
				if (commentflag && checkfound)
				{
					fsetpos(outfile,&cpos);
					fprintf(outfile,"%05u",crc);
					sprintf(pstr,"Checksum : %05u\n",crc);
					pstring(pstr,0);
				}
				fclose(outfile);
			}
			return(nrbosses);
		}
		if (buffer[0]==';')
		{                                   /* test for and skip comment line */
			if (commentflag)
			{
				if (firstline)
				{
					i=0;
					while (buffer[i] && buffer[i] != ':') i++;
					if (buffer[i]==':') { buffer[i+2]=0; checkfound=TRUE; }

					if (mergeflag)
					{
						fprintf(outfile,"%s",buffer);
						fgetpos(outfile,&cpos);				/* remember pos of checksum */
						if (checkfound) fprintf(outfile,"00000\n");
						getdate(&datum);
						i=datum.da_mon; if (i<1 || i>12) i=0;
						sprintf(buffer,";S Pointlist %s inserted by PCMERGE %s on %d-%s-%d\n",
								pointlist,VERSION,datum.da_day,monthn[i],datum.da_year);
						crc=Writeline(buffer,outfile,crc);
						firstline=FALSE;
					}
				}
				else
				if (mergeflag) crc=Writeline(buffer,outfile,crc);
			}
			continue;
		}
		hostflag=zoneflag=FALSE;
		an++;		/* bump nodes processed */	/* first field of line to status */
		if (!splitentry(buffer,status,bnumber,bossysname,bossloc,bossname
														,bossnumber,bossbaud,bossflags))
															Error(14,0,0,0,0,TRUE);
		sscanf(bnumber,"%u",&nmbr);
		if (usebossname)
		{
			if (usebossname & 0x08) strcpy(sysname,bossname);
			else strcpy(sysname,bossysname);
		}
		if (usebossloc) strcpy(location,bossloc);
		if (usebossfone)
		{
			if (!(usebossfone & 0x02)) strcpy(pointphonenr,bossnumber);
			strcpy(baud,bossbaud);
			strcat(baud,",");
			strcat(baud,bossflags);
		}
		if (stricmp(status,"POINT")==0)
		{
			if (mergeflag) crc=Writeline(buffer,outfile,crc);
			continue;
		}
		if (stricmp(status,"ZONE")==0)
		{
			zoneflag=hostflag=TRUE;
			zonenr=nmbr;
			netnr=zonenr;
			nodenr=0;
			p=z=0;
			while (zonenr>p_list[index[z]].zone)		/* hunt for the zone */
			{
				z++;
				if (z>=maxpoints) {z=-1; break;}	/* if no points found for zone */
			}
			m=n=z;
		}
		if (z<0)
		{													  /* if no points for this zone */
			if (mergeflag) crc=Writeline(buffer,outfile,crc); /* write line to output and */
			continue;                             /* skip the rest */
		}
		if (stricmp(status,"REGION")==0 || stricmp(status,"HOST")==0 || zoneflag)
		{
			hostflag=TRUE;
			m=n=z;                              /* point to start of zone */
			p=0;
			netnr=nmbr;
			nodenr=0;
			while (netnr>p_list[index[n]].net)			/* hunt for the net */
			{
				n++;
				if (n>=maxpoints || zonenr!=p_list[index[n]].zone)
				{
					n=-1;
					break;								/* if no points found for net */
				}
			}
		}
		if (!(hostflag || zoneflag))	nodenr=nmbr;
		if (mergeflag) crc=Writeline(buffer,outfile,crc);  /* write line to outfile */
		if (n<0) continue;						/* skip if no points in net */
		m=n;												/* point to current net */
		while (nodenr>p_list[index[m]].node)   /* hunt for any points */
		{
			m++;
			if (m>=maxpoints || netnr!=p_list[index[m]].net ||
						zonenr!=p_list[index[m]].zone)
			{
				m=-1;
				break;									/* if no points found for net */
			}
		}
		if (m<0) continue;
		bossfound=FALSE;
		p=0;
		if (usebossname & 0x10) bracktype=2; else bracktype=3;
		if (usebossname & 0x02) brackup(sysname,sizeof(sysname),bracktype); /* brackets around it */
		if (usebossloc & 0x10) bracktype=2; else bracktype=3;
		if (usebossloc & 0x02) brackup(location,sizeof(location),bracktype);
		if (usebossfone)
		{
			if (usebossfone & 0x02) sprintf(pstr,"%s,%s",bossbaud,bossflags);
			else 	sprintf(pstr,"%s,%s,%s",pointphonenr,bossbaud,bossflags);
		}
		while (nodenr==p_list[index[m+p]].node && netnr==p_list[index[m+p]].net
				&& zonenr==p_list[index[m+p]].zone )    /* write point entry to output */
		{
			if ((p_list[index[m+p]].flags & 0x01)==0)		/* skip deleted entrys */
			{
				bossfound=TRUE;
				if (mergeflag)
				{
					if (usepointname) strcpY(sysname,p_list[index[m+p]].namep);
					if (usebossname & 0x04) sprintf(pstr,"Point_%d_of_",p_list[index[m+p]].point);
					else pstr[0]=0;
					if (usebossname & 0x10) brackup(pstr,sizeof(pstr),1);
					bossin[0]=0;
					if (usebossloc & 0x04) strcpy(bossin,"Boss_in_");
					if (usebossloc & 0x10) brackup(bossin,sizeof(bossin),1);
					sprintf(buffer,"Point,%u,%s%s,%s%s,%s,%s,%s\n",p_list[index[m+p]].point
					,pstr,sysname,bossin,location,p_list[index[m+p]].namep,pointphonenr,baud);
					crc=Writeline(buffer,outfile,crc);
				}
				p_list[index[m+p]].flags |= 0x80; /* mark as handled */
				ap++;                               /* bump points processed */
			}
			if (!toomanybosses) p_list[index[m+p]].bossnr=nextboss; /* mark the number of his boss */
			if (usebossname || usebossloc || usebossfone)
			{
				bosspos[nextboss]=bpos;
			}
			p++;                                /* next point in array */
		}
		if (bossfound)
		{
			if (nextboss<aantbosses-1) nextboss++;
			else
			{
				if (!toomanybosses)
				{

					if (fdflag) { putch('\a');	Error(7,0,0,0,0,FALSE); }
					toomanybosses=TRUE;
				}
			}
			nrbosses++;
		}
		continue;
	}
	/*	return(nrbosses); */
}

/* find an input list */

void findlist(char *fbuf, char *inlist, char *defname, char *defext,int ern)
{
	int i;
	char fffdrv[MAXDRIVE], fffpath[MAXPATH], fffname[MAXFILE], fffext[MAXEXT];

	strcpy(fbuf,inlist);
	i=fnsplit(fbuf,fffdrv,fffpath,fffname,fffext);
	if ((i & FILENAME) == 0)
	{
		strcpy(fffname,defname);
		strcpy(fffext,defext);
	}
	else if ((i & EXTENSION) == 0) strcpy(fffext,".???");
	fnmerge(fbuf,fffdrv,fffpath,fffname,fffext);
	if (directory(fbuf,inlist)==-1) Error(ern,0,0,0,0,TRUE);
	fnsplit(inlist,NULL,NULL,fffname,fffext);
	fnmerge(fbuf,fffdrv,fffpath,fffname,fffext);
	strcpy(inlist,fbuf);
}

/* directory searching */

int directory(char *match, char *fname)
{
   #ifdef __WATCOMC__
      struct find_t infoblk;
   #else
      struct ffblk infoblk;
      char far *dtasave;
   #endif

	char namef[MAXFILE],extf[MAXEXT],extf1[MAXEXT];
	int i,j,ret=-1;


   #ifndef __WATCOMC__
     dtasave = getdta();           /* Save het org. disk transfer adres. */
   #endif

	if (findfirst(match,&infoblk,0)==0)
	{
      #ifndef __WATCOMC__
		strcpy(fname,infoblk.ff_name);
      #else
      strcpy(fname,infoblk.name);
      #endif

		fnsplit(fname,NULL,NULL,namef,extf);
		strcpy(extf1,extf);
		if (isdigit(extf[1]) && isdigit(extf[2]) && isdigit(extf[3])
							&& extf[1]!='9') sscanf(&extf[1],"%d",&i); else i=0;

		while (findnext(&infoblk)==0)
		{
         #ifndef __WATCOMC__
           fnsplit(infoblk.ff_name,NULL,NULL,namef,extf);
         #else
           fnsplit(infoblk.name,NULL,NULL,namef,extf);
         #endif

			if (forbiddenext(extf)) continue;
         if (forbiddenext(extf1))
             #ifndef __WATCOMC__
               strcpy(fname,infoblk.ff_name);
             #else
               strcpy(fname,infoblk.name);
             #endif

			strcpy(extf1,extf);
			if (isdigit(extf[1]) && isdigit(extf[2]) &&
								isdigit(extf[3]) && extf[1]!='9' )
			{
				sscanf(&extf[1],"%d",&j);
				if ( i==0 || (j>i && (j-i)<183) || (j<i && (i-j)>183) )
				{
               #ifndef __WATCOMC__
                 strcpy(fname,infoblk.ff_name);
               #else
                 strcpy(fname,infoblk.name);
               #endif
					i=j;
				}
			}
		}
		fnsplit(fname,NULL,NULL,NULL,extf);
		if (i==999 || stricmp(extf,".BAK")==0) ret=-1; else ret=0;
	}

   #ifndef __WATCOMC__
      setdta(dtasave);              /* org disk transfer adres herstellen */
   #else
      _dos_findclose(&infoblk);
   #endif

	return(ret);
}

/* test if extension is forbidden
	forbidden extesions are BAK, BAT, COM, EXE, Ann, Jnn, Lnn, Rnn, Snn and Znn */

boolean forbiddenext(char *f)
{

	char c;

	if (strlen(f) !=4) return(FALSE);
	if (f[0]!='.') return(TRUE);

	c=toupper(f[1]);

	if (	stricmp(f,".BAK")==0 ||  stricmp(f,".BAT")==0 ||
			stricmp(f,".EXE")==0 ||  stricmp(f,".COM")==0 ||
			(	isdigit(f[2]) && isdigit(f[3]) &&
				(c=='A' || c=='J' || c== 'L' || c=='R' || c=='S' || c=='Z') ) )
			return(TRUE);

	return(FALSE);
}


/* same as fgets except that it advances pointer to EOL
	after coping as many chars as possible  */

char *Fgets(char *s,int n, FILE *fp)
{
	char cc;
	int ii,jj=0;

	for (ii=0;ii<n-1;ii++)
	{
		cc=fgetc(fp);
		if (cc=='\r') continue;
		if (cc=='\n' || cc==EOF) break;
		s[jj++]=cc;
	}
	s[jj]=0;
	while (cc!='\n' && cc!=EOF) cc=fgetc(fp);
	if (cc==EOF && jj==0) return(NULL); else return(s);
}

/* read the config file if present */

int readconfig(char *s, int maxl)
{
	FILE *f;
	int i,n=0,cfghandle;

	cfghandle=open(CONFIGFILE,O_RDONLY|O_DENYNONE);
	if (cfghandle==-1) return(0);

	if ((f=fdopen(cfghandle,"rt"))==NULL) return(0);	/* if no config */

	printf("\nReading config file.\n");
	while (Fgets(s,maxl,f)!=NULL)
	{
		n++;
		if ((i=scanoption(s,maxl))!=0) { fclose(f); return(i*n); }
	}
	fclose(f);
	return(0);
}

/* scan the command line for options and filenames */

int getcommands(int argc, char *argv[])
{
	int i,n;
	char s[128];

	for (i=1;i<argc;i++)
	{
		strcpy(s,argv[i]);
		if ((n=scanoption(s,sizeof(s)-1))!=0) return(n*i);
	}
	return(0);
}

/* Scan an option field. Return 0 on succes, 1 if the option is unknown
	and -1 if the sunstitution of an environment variable fails */

int scanoption(char *s, int maxl)
{
	int i,j,k=0;
	boolean minus,commainbaud=FALSE;
	char c,d,r=0;
	const char sepchar[]={'.','_',' '};

	minus=FALSE; j=0;
	if (s[0]==';' || s[0]=='\r' || s[0]=='\n') return(0);

	if (!substenvirons(s,maxl)) return(-1);

	while ((c=toupper(s[j++]))!=0)
	{
		if (c=='/') continue;
		if (c=='-') { minus=TRUE; continue; }
		d=toupper(s[j]);
		if (d=='=' || d==':') d=toupper(s[++j]);
		switch (c)
		{
			case 'A' :	errabort=!minus; break;
			case 'B'	:	if (d)
							{
								strcpy(baud,&s[j]);
								i=0;
								while (baud[i])
								{
									if (baud[i]==',') { commainbaud=TRUE; break; }
									i++;
								}
								if (!commainbaud) strcat(baud,",");
								if (stricmp(baud,"BOSS,")==0) usebossfone=3;
							}
							break;
			case 'C'	:	commentflag=FALSE; break;
			case 'D' :	if (minus) nameclean=0; else nameclean=-1; break;
			case 'E' :	excludeorphan=!minus; break;
			case 'F'	:	if ((fdflag=!minus)==TRUE) if (d) strcpy(fdlist,&s[j]);
							break;
			case 'G' :  if (d) sscanf(&s[j],"%u",&Defaultzone);
							else Defaultzone=DEFAULTZONE;
							break;
			case 'K'	:  if (minus) privats=-1;
							else
							{
								if (privats<NRPRIVATS-1)
								{
									privats++;
									if (d) Strcpy(pvtlist[privats],&s[j]);
								}	else pwarning=TRUE;
							}
							break;
			case 'L'	:  if (minus) logflag=2; else logflag=TRUE;
							if (d) strcpy(logfile,&s[j]);
							else strcpy(logfile,LOGFILE);
							break;
			case 'M'	:  if (minus) multiflag=2; else multiflag=1; break;
			case 'N'	:	if (d) strcpy(nodelist,&s[j]);
							break;
			case 'O'	:	if ((mergeflag=!minus)==TRUE) if (d) strcpy(outlist,&s[j]);
							break;
			case 'P'	:  if (d) strcpy(pointlist,&s[j]);
							break;
			case 'Q'	:  if (minus) maxeolerrors=0;
							else if (d) sscanf(&s[j],"%d",&maxeolerrors);
							break;
			case 'S'	:  sortflag=!minus;
							if (d) strcpy(spointlist,&s[j]);
							break;
			case 'U'	:  if (usebossfone!=3) usebossfone=0;
							if (d)
							{
								strcpy(pointphonenr,&s[j]);
								if (stricmp(pointphonenr,"BOSS")==0) usebossfone=1;
							}
							else strcpy(pointphonenr,"-U-");
							break;
			case 'X'	:  usebossname=usepointname=0;
							if (minus) sysname[0]=0; else
							{
								if (d) strcpy(sysname,&s[j]);
								if (stricmp(sysname,"BOSS")==0) usebossname=0x01;
								if (stricmp(sysname,"[BOSS]")==0) usebossname=0x03;
								if (stricmp(sysname,".BOSS")==0) usebossname=0x05;
								if (stricmp(sysname,".[BOSS]")==0) usebossname=0x07;
								if (stricmp(sysname,"[.BOSS]")==0) usebossname=0x17;
								if (stricmp(sysname,"BOSSNAME")==0) usebossname=0x09;
								if (stricmp(sysname,"[BOSSNAME]")==0) usebossname=0x0B;
								if (stricmp(sysname,".BOSSNAME")==0) usebossname=0x0D;
								if (stricmp(sysname,".[BOSSNAME]")==0) usebossname=0x0F;
								if (stricmp(sysname,"[.BOSSNAME]")==0) usebossname=0x1F;
								if (stricmp(sysname,"POINT")==0) usepointname=TRUE;
							}
							break;
			case 'Y'	:  usebossloc=0;
							if (minus) location[0]=0; else
							{
								if (d) strcpy(location,&s[j]);
								if (stricmp(location,"BOSS")==0) usebossloc=0x01;
								if (stricmp(location,"[BOSS]")==0) usebossloc=0x03;
								if (stricmp(location,".BOSS")==0) usebossloc=0x05;
								if (stricmp(location,".[BOSS]")==0) usebossloc=0x07;
								if (stricmp(location,"[.BOSS]")==0) usebossloc=0x17;
							}
							break;
			case 'Z'	:	zapflag=!minus; break;
			case '*'	:	if (!multiflag) multiflag=1;
							if (!logflag) { logflag=2; strcpy(logfile,LOGFILE); }
							if (minus) npcflag=2; else npcflag=TRUE;
							if (d) sscanf(&s[j],"%d",&k);
							if (k>2) { r=1; break; }
							separator=sepchar[k];
							break;
			default	:	r=1; break;
		}
		break;
	}
	return(r);
}

/* Search for and substitute %strings% by the corresponding environment
	variables. Returns TRUE if succesfull. */

boolean substenvirons(char *s, int max)
{
	char buf1[30], buf[256], *c;
	int i,j,k;

	buf1[0]=0;
	i=0; k=0;

	while (TRUE)
	{
		while(s[i] && s[i]!='%') { buf[k]=s[i]; i++; k++; }

		if (s[i]==0)
		{
			buf[k]=0; j=0;
			while (j<max-1 && buf[j]) { s[j]=buf[j]; j++; }
			s[j]=0;
			break;
		}
		i++;
		buf1[0]=0; j=0;
		while (s[i] && s[i]!='%')
		{
			buf1[j]=toupper(s[i]);
			i++; j++;
		}
		buf1[j]=0;

		if (s[i]=='%') i++; else return(FALSE);

		if ((c=getenv(buf1))==NULL) return(FALSE);
		j=0;
		while (c[j] && k<max-2) { buf[k]=c[j]; k++; j++; }
	}
	return(TRUE);
}

/* timestamp to log file */

void timestamp(FILE *f, int s)
{
	const char ind[]={'#','!','?','$','*',':','+','-','%','~','='};

   #ifdef __WATCOMC__

     struct dostime_t d;

     _dos_gettime(&d);
     fprintf(f,"%c %2d:%02d:%02d  ", ind[s], d.hour, d.minute, d.second);

   #else

     struct time ftime;

     gettime(&ftime);
	  fprintf(f,"%c %2d:%02d:%02d  ",ind[s],ftime.ti_hour,ftime.ti_min,ftime.ti_sec);

   #endif
}

/* Error handler */

void Error(int er, int ln, FILE *f, fpos_t pos, int cpos, boolean errab)
{
	char c=0;
	int i;
	char line[80];
	char buf[6];

	line[0]=0;
	if (ln>0) sprintf(line,"Line number %d : ",ln);

	switch(er)
	{
		case 0	:	strcat(line,"Warning: default ZONE (");
						sprintf(buf,"%d",Defaultzone);
						strcat(line,buf);
						strcat(line,") used.");
						break;
		case 1	:  strcat(line,"Warning: Comma or '");
						sprintf(buf,"%c",separator);
						strcat(line,buf);
						strcat(line,"' expected.");
						break;
		case 2	:  strcat(line,"Warning: More than 20 private lists, rest ignored.");
						break;
		case 3	:	strcat(line,"Warning: First name missing."); break;
		case 4	:	strcat(line,"Error opening logfile.");break;
		case 5	:	strcat(line,"Error in number."); break;
		case 6	:	strcat(line,"Unexpected end of line.");
						eolerrors++; break;
		case 7	:	strcat(line,"Out of memory for bossdata!, rest not stored!");
						break;
		case 10	:	strcat(line,"Can't open output file."); break;
		case 12	:	strcat(line,"Can't open nodelist segment."); break;
		case 13	:	strcat(line,"Can't open merged output file."); break;
		case 14	:	strcat(line,"Error reading nodelist segment."); break;
		case 15	:	strcat(line,"Error opening input point list."); break;
		case 16	:	strcat(line,"Can't find a pointlist for input.");break;
		case 17	:	strcat(line,"Cant find a nodelist segment for input.");break;
		case 21	:	strcat(line,"Can't find private pointlist.");break;
		case 22	:	strcat(line,"Error opening private pointlist."); break;
		case 98  :  strcat(line,"User abort or timeout."); break;
		case 99	:	strcat(line,"Not enough memory."); break;
		case 255	:	strcat(line,"List is probably not a Fido_User list."); break;
		default	:	;
	}
	sprintf(pstr,"%s\n",line);
	pstring(pstr,2);
	if (ln>0)
	{
		fsetpos(f,&pos);				/* go to start of offending line */
		i=0;
		while(c != '\n' && c!=EOF && i<(sizeof(line)/sizeof(char))-1)
		{
			c=fgetc(f);             /* and copy to buffer */
			if (isprint(c)) line[i++]=c; else line[i++]=' ';
		}
		line[i]=0;
		sprintf(pstr,"%s\n",line);
		pstring(pstr,2);
		for (i=0;i<cpos-1 && i<sizeof(line)-2;i++) line[i]=' ';
		line[i++]='^';
		line[i]=0;
		sprintf(pstr,"%s\n",line);
		pstring(pstr,0);
	}
	if (errabort || errab)
	{
		sprintf(pstr,"Conversion aborted.\n");
		pstring(pstr,2);
		if (logflag) fclose(log);
		exit(er);
	}
	else
	{
		if (er>3)
		{
			if (ln>0)
			{
				sprintf(pstr,"Offending line skipped\n");
				pstring(pstr,2);
				nrofskips++;
			}
			if (er>errornumber) errornumber=er;
		}
		while (c!='\n' && c!=EOF) c=fgetc(f);			/* skip to eol */
		if (er==6 && maxeolerrors && eolerrors>=maxeolerrors)
																	Error(255,0,0,0,0,TRUE);
		return;
	}
}

/* error reading private pointlist */

void pvterror(int l,char *b)
{
	longstring s;

	sprintf(s,"Line %d : Error in private list.",l);
	sprintf(pstr,"%s\n",s);
	pstring(pstr,2);
	b[65]=0;
	sprintf(pstr,"%s\n",b);
	pstring(pstr,0);
	if (errornumber==0) errornumber=1;
}

/* string output to screen and to log */

void pstring(char *st,int n)
{
	printf("%s",st);
	if (logflag)
	{
		if (n>=0) timestamp(log,n);
		fprintf(log,"%s",st);
	}
}

/* print short helptext */

void help(void)
{
	printf("\n");
	printf("\nPCMERGE  [a] [b=baudrate] [-c] [[-]d] [e] [[-]f=fdlist] [g=defaultzone]");
	printf("\n         [k[=killlist]] [[-]l[=logfile]] [[-]m] [n=nodelist]");
	printf("\n         [-o|o=outputlist] [p=pointlist] [[-]q[=maxeolerrors]] [s[=sortfile]]");
	printf("\n         [[-]u][[-]x[=systemname]][[-]y[=location]] [z] [[-]*[=0|1|2]]");
	printf("\n");
}


/* Calculate the CRC.
	On entry a character is in ch.
	This is a cumulative process, all chars
	CRC's are taken as a running total.
	To start a new sequence, first clear crc.
   - Inspired by PA0MMV's REMOTE! -  */


unsigned int chcrc(unsigned char ch, unsigned int crc)

{
	unsigned char b,crch;

	b=(crc>>8)^ch;
	crch=(crc&0xFF)^CRC1[b];
	return((crch<<8)+CRC2[b]);
}

unsigned int Writeline(char *b, FILE *f, unsigned int ch)
{
	int i=0;

	while (b[i] != 0)
	{
		if (b[i]=='\n') ch=chcrc('\r',ch);
		ch=chcrc(b[i],ch);
		i++;
	}
	fputs(b,f);
	return(ch);
}

/* split a nodelist entry in it components */

boolean splitentry(char *b, char *st,char *nr,char *bsn,char *bl
												,char *bna,char *bnr,char *bd,char *bf)
{
	int i=0,j,k;
	char c=0;
	char *p[8];
	int userflag=0;

	p[0]=st; p[1]=nr; p[2]=bsn; p[3]=bl;
	p[4]=bna; p[5]=bnr; p[6]=bd; p[7]=bf;

	for (j=0;j<8;j++) p[j][0]=0;			/* all fields empty */
	if (b[0]==';') return(TRUE);
	for (j=0;j<7;j++)
	{
		k=0;
		while ((c=b[i])!=',' && c!='\n' && c!=' ' && c!=EOF && c!=0)
		{
			if (k<(sizeof(string)-1)) { p[j][k]=c; k++; }
			i++;
		}
		p[j][k]=0;
		if (c==',') i++; else break;
	}
	if (j!=7) return(FALSE);
	k=0;
	while ((c=b[i])!='\n' && c!=' ' && c!=EOF)
	{
		if (k<(sizeof(longstring)-1))
		{
			if (!userflag && c==',' && b[i+1]=='U') userflag=k+1;
			p[7][k]=c; k++;
		}
		i++;
	}
	p[7][k]=0;
	if (userflag) remove_starC(p[7],userflag);
	return(TRUE);
}

/* remove NC, NEC, REC, ZEC, IEC form user flag field */

void remove_starC(char *us, int uf)
{
	int i,j,k,l,m,n;
	char flag[33];
	static char starc[6][4]={"NC","NEC","REC","ZEC","IC","IEC"};

	i=uf+1;
	if (us[i]==',') i++;
	while (us[i])
	{
		k=i;
		flag[0]=0; j=0;
		while (us[i] && us[i]!=',' && j<sizeof(flag)-1) {flag[j]=us[i];i++;j++;}
		flag[j]=0;
		if (us[i]==',') i++;
		for (n=0;n<sizeof(starc)/sizeof(starc[0]);n++)
		{
			if (strcmp(flag,starc[n])==0)
			{
				l=i;
				m=k;
				do {us[m]=us[l]; m++; l++; } while (us[m]);
			}
		}
	}
	if (us[uf+1]==0 || (us[uf+1]==',' && us[uf+2]==0)) us[uf-1]=0; /*remove empty user flag */
}

/* put square brackets around a string */

void brackup(char *s, unsigned int size, byte type)

/* type==0 : do nothing
	type==1 : square bracket at start of string
	type==2 : square bracket at end of string
	type==3 : both		*/

{
	int i,k;
	char c,d;

	if (size<3 || type==0) return;
	c=s[0];
	if (type & 0x02) k=2; else k=1;
	if (type & 0x01)
	{
		s[0]='[';
		if (type & 0x02) k=2;
		for (i=1;i<size-k;i++)
		{
			d=c;
			c=s[i];
			s[i]=d;
			if (d==0) break;
		}
	}
	else i=min(strlen(s),size-k);
	if (type & 0x02) { s[i]=']'; s[i+1]=0; } else s[i]=0;
}

/* See if a string is an empty line - nothing but spaces */

boolean isemptyline(char *s)
{
	int i=0;

	while (s[i])
	{
		if (isgraph(s[i])) return(FALSE);
		i++;
	}
	return(TRUE);
}

/* that's it folks */


/* *************************************************** */
/* Some compatibility functions for Watcom and/or OS/2 */
/* *************************************************** */

#if defined (__WATCOMC__) && defined (__OS2__)

void getdate(struct date *d)
{
   struct dosdate_t dosdate;

   _dos_getdate(&dosdate);

   d->da_day   = dosdate.day;
	d->da_mon   = dosdate.month;
   d->da_year  = dosdate.year;
}


int fnsplit(const char *path, char *drive, char *dir, char *name, char *ext)
{
   int retval = 0;

   _splitpath(path, drive, dir, name, ext);

   if(drive && *drive != '\0') retval |= DRIVE;
   if(dir   && *dir   != '\0') retval |= DIRECTORY;
   if(name  && *name  != '\0') retval |= FILENAME;
   if(ext   && *ext   != '\0') retval |= EXTENSION;

   return retval;
}

#endif
